#------------------------------------------------------------------------------
# Libraries
#------------------------------------------------------------------------------
# Standard
import pandas as pd
import numpy as np
import bodyguard as bg

from difflib import SequenceMatcher
#------------------------------------------------------------------------------
# Main
#------------------------------------------------------------------------------
class TextMatcher(object):
    """
    Match text
    """
    # -------------------------------------------------------------------------
    # Constructor function
    # -------------------------------------------------------------------------
    def __init__(self,
                 method="Ratcliff-Obershelp",
                 speed="slow",
                 ):
        self.method = method
        self.speed = speed
        
        if not bg.tools.isin(a=self.method, b=self.METHOD_OPT):
            bg.exceptions.WrongInputException(input_name="method",
                                              provided_input=self.method,
                                              allowed_inputs=self.METHOD_OPT)
    # -------------------------------------------------------------------------
    # Class variables
    # -------------------------------------------------------------------------
    METHOD_OPT = ["Ratcliff-Obershelp"]
    SPEED_OPT = ["slow", "quick", "real_quick"]
    
    # -------------------------------------------------------------------------
    # Private functions
    # -------------------------------------------------------------------------
    def _check_a(self, a):
        if not isinstance(a,str):
            raise bg.exceptions.WrongInputTypeException(input_name="a",
                                                        provided_input=a,
                                                        allowed_inputs=str)

    def _check_b(self, b):
        if not isinstance(b,list):
            raise bg.exceptions.WrongInputTypeException(input_name="b",
                                                        provided_input=b,
                                                        allowed_inputs=list)

    
    def _ratcliff_obershelp(self,a,b,speed="slow",isjunk=None,autojunk=True):
        
        
        if speed=="slow":        
            seq_similarity = SequenceMatcher(isjunk=isjunk,
                                             a=a,
                                             b=b,
                                             autojunk=autojunk).ratio()
        elif speed=="quick":
            seq_similarity = SequenceMatcher(isjunk=isjunk,
                                             a=a,
                                             b=b,
                                             autojunk=autojunk).quick_ratio()
        elif speed=="real_quick":
            seq_similarity = SequenceMatcher(isjunk=isjunk,
                                             a=a,
                                             b=b,
                                             autojunk=autojunk).real_quick_ratio()
        else:
            raise bg.exceptions.WrongInputException(input_name="speed",
                                                    provided_input=speed,
                                                    allowed_inputs=self.SPEED_OPT)
            
        return seq_similarity
            
    # -------------------------------------------------------------------------
    # Public functions
    # -------------------------------------------------------------------------
    def compute_similarity(self,a,b,**kwargs):
        """
        Compute similarity between string and each element in list
        """
        # Sanity checks
        self._check_a(a=a)
        self._check_b(b=b)
        
        if self.method=="Ratcliff-Obershelp":
            similarities = [self._ratcliff_obershelp(a=a,
                                                     b=x,
                                                     speed=self.speed,
                                                     **kwargs) for x in b]
            
            
        return similarities
            
            
        
# from diff_match_patch import diff_match_patch
# P_temp[P_temp.index.get_level_values(key) == code_from] = [diff.diff_levenshtein(diffs=diff.diff_main(text1=code_from, text2=k)) / max(len(code_from),len(k)) for k in key_to]
        


#------------------------------------------------------------------------------
# Libraries
#------------------------------------------------------------------------------
# Standard
import os
import pandas as pd
import numpy as np
import bodyguard as bg

# Using SentenceTransformers NLP library
from sentence_transformers import SentenceTransformer

# Using HuggingFace NLP library
from transformers import AutoTokenizer, AutoModel
import torch

# From this library
from ..utils.convert import convert_dict_to_df, convert_df_to_dict
from ..utils.distance import normalize_by_norm
#------------------------------------------------------------------------------
# Main
#------------------------------------------------------------------------------
class DocumentEmbedder(object):
    """
    Embedding of documents
    
    We rely on two major libraries to embed documents, namely https://github.com/UKPLab/sentence-transformers and https://huggingface.co/transformers/
    """
    # -------------------------------------------------------------------------
    # Constructor function
    # -------------------------------------------------------------------------
    def __init__(self,
                 normalize=True,
                 verbose=True,
                 ):
        self.normalize = normalize
        self.verbose = verbose
        
    # -------------------------------------------------------------------------
    # Class variables
    # -------------------------------------------------------------------------
    STR_USING_SAVED_MODEL = "Using saved model: '{0}'"
    STR_DOWNLOADING_MODEL = "Downloading model: '{0}'"    
    STR_DOCUMENTS_NEED_EMBEDDING_INPUT = "Embeddings from model '{0}' needed"

    # -------------------------------------------------------------------------
    # Private functions
    # -------------------------------------------------------------------------
    #Mean Pooling - Take attention mask into account for correct averaging
    def _mean_pooling(self, model_output, attention_mask):
        """
        See example here: https://huggingface.co/sentence-transformers/all-roberta-large-v1
        """
        token_embeddings = model_output[0] #First element of model_output contains all token embeddings
        input_mask_expanded = attention_mask.unsqueeze(-1).expand(token_embeddings.size()).float()
        sum_embeddings = torch.sum(token_embeddings * input_mask_expanded, 1)
        sum_mask = torch.clamp(input_mask_expanded.sum(1), min=1e-9)
    
        return sum_embeddings / sum_mask

    def _check_embeddings_existence(self, documents, model_name_or_path):
        """Check which documents that need to be embeddings versus which documents that have already been embedded"""
        
        # Check if embeddings exist using this particular method
        if hasattr(self, model_name_or_path+"-embeddings"):
            if self.verbose:
                print(f"Embeddings from model '{model_name_or_path}' found")
            # Load existing embeddings
            documents_embeddings = getattr(self, model_name_or_path+"-embeddings")
            
            # Check which documents have not already been embedded
            documents_not_embedded = []
            for doc in documents:
                if documents_embeddings.get(doc) is None:
                    documents_not_embedded.append(doc)                
        else:
            setattr(self, model_name_or_path+"-embeddings", {})
            documents_not_embedded = documents
        
        return documents_not_embedded

    
    def _extract_existing_embeddings(self, documents, model_name_or_path):
        """Extract those embeddings that have already been embedded"""
        # Extract relevant document embeddings
        embeddings = {k:getattr(self, model_name_or_path+"-embeddings")[k] for k in documents}
        
        return embeddings


    def _embed_documents(self,
                         documents,
                         model_name_or_path="all-roberta-large-v1"):
        # ---------------------------------------------------------------------
        # Check documents have been embedded before
        # ---------------------------------------------------------------------
        documents_not_embedded = self._check_embeddings_existence(documents=documents,
                                                                  model_name_or_path=model_name_or_path)

        if bool(documents_not_embedded):
            # We have assessed that some documents have not been embedded, hence we construct embeddings and save them!
            if self.verbose:
                print(self.STR_DOCUMENTS_NEED_EMBEDDING_INPUT.format(model_name_or_path))
                
            # -----------------------------------------------------------------
            # Check if model is already loaded
            # -----------------------------------------------------------------
            if hasattr(self, model_name_or_path+"-model"):
                if self.verbose:
                    print(self.STR_USING_SAVED_MODEL.format(model_name_or_path))
                    
                pretrained_model = getattr(self, model_name_or_path+"-model")
                tokenizer = getattr(self, model_name_or_path+"-tokenizer")
                
            else:
                if self.verbose:
                    print(self.STR_DOWNLOADING_MODEL.format(model_name_or_path))
    
                try:
                    # Load model via SentenceTransformer
                    pretrained_model = SentenceTransformer(model_name_or_path=model_name_or_path)
                    tokenizer = None
                    self.loaded_via = "SentenceTransformer"
                except:                
                    try:
                        # Load model via HuggingFace
                        pretrained_model = AutoModel.from_pretrained(model_name_or_path)
                        tokenizer = AutoTokenizer.from_pretrained(model_name_or_path)
                        self.loaded_via = "HuggingFace"
                    except OSError as err:
                        raise Exception(f"Could not load '{model_name_or_path}'. \nWhile trying, this error occured: \n\n{err}")
                        
                # Set tokenizer and pretrained model as attributes under the model name.
                setattr(self, model_name_or_path+"-model", pretrained_model)
                setattr(self, model_name_or_path+"-tokenizer", tokenizer)        
                
            # -----------------------------------------------------------------
            # Embed
            # -----------------------------------------------------------------
            # Maximum sentence length
            max_seq_length = max([256, max([len(d.split()) for d in documents_not_embedded])])    
            
            if self.loaded_via=="SentenceTransformer":
                            
                # Update maximum sequence length
                pretrained_model.max_seq_length = max([pretrained_model.max_seq_length, max_seq_length])            
    
                # Construct embedding
                raw_embeddings = pretrained_model.encode(documents_not_embedded,
                                                         batch_size=32,
                                                         show_progress_bar=False,
                                                         output_value="sentence_embedding",
                                                         convert_to_numpy=True,
                                                         convert_to_tensor=False,
                                                         device=None,
                                                         normalize_embeddings=True)
                            
            elif self.loaded_via=="HuggingFace":
    
                # Tokenize sentences
                encoded_input = tokenizer(text=documents_not_embedded,
                                          padding=True,
                                          truncation=True,
                                          max_length=max_seq_length,
                                          return_tensors='pt')
            
                # Compute token embeddings
                with torch.no_grad():
                    model_output = pretrained_model(**encoded_input)
            
                # Perform pooling (here, mean pooling)
                raw_embeddings = self._mean_pooling(model_output=model_output, attention_mask=encoded_input['attention_mask'])                
    
                # Convert tensor to numpy array
                raw_embeddings = raw_embeddings.numpy()
    
            # -----------------------------------------------------------------
            # Finalize
            # -----------------------------------------------------------------
                    
            # Pre-allocate
            embeddings = {}
            
            for i,doc in enumerate(documents_not_embedded):
                embeddings[doc] = raw_embeddings[i,:].reshape(-1,)
                
            # Update existing document embeddings
            setattr(self, model_name_or_path+"-embeddings", {**getattr(self, model_name_or_path+"-embeddings"),
                                                             **embeddings}
                    )            
            
            
        # Extract embeddings
        embeddings_out = self._extract_existing_embeddings(documents=documents,
                                                           model_name_or_path=model_name_or_path)
        
        return embeddings_out
    # -------------------------------------------------------------------------
    # Public functions
    # -------------------------------------------------------------------------
    def show_available_models(self):
        """Print available models via SentenceTransformers"""
        
        # Last last updated
        DATE_LAST_UPDATED = "February 3, 2022"
        
        # Models
        AVAILABLE_MODELS_VIA_SENTENCETRANSFORMERS = ["all-roberta-large-v1",
                                                     "all-mpnet-base-v1",
                                                     "all-mpnet-base-v2"
                                                     "all-MiniLM-L12-v1",
                                                     "all-distilroberta-v1",
                                                     "all-MiniLM-L12-v2",
                                                     "all-MiniLM-L6-v2",
                                                     "all-MiniLM-L6-v1",
                                                     "paraphrase-mpnet-base-v2",
                                                     "multi-qa-mpnet-base-dot-v1",
                                                     "multi-qa-distilbert-dot-v1",
                                                     "multi-qa-mpnet-base-cos-v1",
                                                     "paraphrase-distilroberta-base-v2",
                                                     "paraphrase-TinyBERT-L6-v2",
                                                     "paraphrase-MiniLM-L12-v2",
                                                     "multi-qa-distilbert-cos-v1",
                                                     "paraphrase-multilingual-mpnet-base-v2",
                                                     "paraphrase-MiniLM-L6-v2",
                                                     "paraphrase-albert-small-v2",
                                                     "multi-qa-MiniLM-L6-cos-v1",
                                                     "paraphrase-multilingual-MiniLM-L12-v2",
                                                     "multi-qa-MiniLM-L6-dot-v1",
                                                     "msmarco-bert-base-dot-v5",
                                                     "msmarco-distilbert-base-tas-b",
                                                     "paraphrase-MiniLM-L3-v2",
                                                     "msmarco-distilbert-dot-v5",
                                                     "distiluse-base-multilingual-cased-v1",
                                                     "distiluse-base-multilingual-cased-v2",
                                                     "average_word_embeddings_komninos",
                                                     "average_word_embeddings_glove.6B.300d",
                                                     ]
        
        print(f"""As of out {DATE_LAST_UPDATED}, these models are available through SentenceTransformers:
              
              {AVAILABLE_MODELS_VIA_SENTENCETRANSFORMERS}
              """
              )
        return AVAILABLE_MODELS_VIA_SENTENCETRANSFORMERS


    def embed_documents(self,
                        documents,
                        return_type="df",
                        model_name_or_path="all-roberta-large-v1"):
        RETURN_TYPE_OPT = ["dict", "df"]
        
        # Check documents
        if isinstance(documents, list):
            pass
        elif isinstance(documents, str):
            documents = [documents]
        else:
            raise bg.exceptions.WrongInputException(input_name="documents",
                                                    provided_input=documents,
                                                    allowed_inputs=["str", "list"])
            
        # Embed documents
        embeddings = self._embed_documents(documents=documents,
                                           model_name_or_path=model_name_or_path)
                    
        # Normalize
        if self.normalize:
            if isinstance(embeddings, dict):             
                embeddings = {k: normalize_by_norm(x=v,norm="L2") for k,v in embeddings.items()}
                
        # Convert to df to change column type              
        embeddings = convert_dict_to_df(x=embeddings)
        
        # Enforce columns to be strings
        embeddings.columns = embeddings.columns.astype(str)
                
        if return_type=="df":
            pass
        elif return_type=="dict":
            embeddings = convert_df_to_dict(x=embeddings)    
        else:
            raise bg.exceptions.WrongInputException(input_name="return_type",
                                                    provided_input=return_type,
                                                    allowed_inputs=RETURN_TYPE_OPT)
        
        return embeddings
    







#------------------------------------------------------------------------------
# Libraries
#------------------------------------------------------------------------------
# Standard
import pandas as pd
import bodyguard as bg
import re

from .matching.text_matching import TextMatcher
from .matching.numeric_matching import NumericMatcher

#------------------------------------------------------------------------------
# Main
#------------------------------------------------------------------------------
class AutoCrosswalk(object):
    """
    This class crosswalks between two documents 
    """
    # -------------------------------------------------------------------------
    # Constructor function
    # -------------------------------------------------------------------------
    def __init__(self,
                 n_best_match=2,
                 prioritize_exact_match=True,
                 enforce_completeness=True,
                 verbose=False):
        self.n_best_match = n_best_match
        self.prioritize_exact_match=prioritize_exact_match
        self.enforce_completeness=enforce_completeness
        self.verbose = verbose
        
        # Pre-set variables
        self.has_all_keys = False
        self.unique_from = []
        self.unique_to = []
    # -------------------------------------------------------------------------
    # Class variables
    # -------------------------------------------------------------------------
    NUMERIC_TYPE_OPT = ["continuous", "categorical"]
    TOL = 0.001

    # -------------------------------------------------------------------------
    # Very private functions
    # -------------------------------------------------------------------------
    def __check_numeric_keys(self, numeric_key, numeric_type):
        """
        Sanity check numeric keys and type
        """
        
        if numeric_key is None:
            pass
        elif not isinstance(numeric_key, list):
            raise Exception(f"When 'numeric_key' is not None, it must be an instance of list. Currently, it is {type(numeric_key)}")
        else:
            if not bg.tools.isin(a=numeric_type, b=self.NUMERIC_TYPE_OPT):
                raise bg.exceptions.WrongInputException(input_name="numeric_type",
                                                        provided_input=numeric_type,
                                                        allowed_inputs=self.NUMERIC_TYPE_OPT)

    def __check_text_keys(self, text_key):
        """
        Sanity check text keys
        """
        if text_key is None:
            pass
        elif not isinstance(text_key, list):
            raise Exception(f"When 'text_key' is not None, it must be an instance of list. Currently, it is {type(text_key)}")
    
    def __fix_values(self,v):
        if isinstance(v, list):
            pass
        elif isinstance(v, str):
            v = [v]
        else:
            raise bg.exceptions.WrongInputTypeException(input_name="v",
                                                        provided_input=v,
                                                        allowed_inputs=[list, str])
        
        return v
    
    
    def __fix_by(self,by,cols):        
        if not isinstance(cols,list):
            raise bg.exceptions.WrongInputTypeException(input_name="cols",
                                                        provided_input=cols,
                                                        allowed_inputs=list)
        
        if by is None:
            by = cols
        elif isinstance(by, list):
            
            # Exclude elem from cols
            by = [x for x in by if not bg.tools.isin(a=x, b=cols)]
            
            by += cols
        elif isinstance(by, str):
            
            if bg.tools.isin(a=by, b=cols):
                by = cols
            else:            
                by = cols+[by]
        else:
            raise bg.exceptions.WrongInputTypeException(input_name="by",
                                                        provided_input=by,
                                                        allowed_inputs=[None, list, str])
        return by
    
    # -------------------------------------------------------------------------
    # Private functions
    # -------------------------------------------------------------------------
    def _check_keys(self, numeric_key, numeric_type, text_key):
        """
        Sanity check all keys
        """
        self.__check_numeric_keys(numeric_key=numeric_key, numeric_type=numeric_type)
        self.__check_text_keys(text_key=text_key)
        
        if (numeric_key is None) and (text_key is None):
            raise Exception("Both 'numeric_key' and 'text_key' cannot be None simultaneously")
        elif numeric_key is None:
            numeric_key = []
        elif text_key is None:
            text_key = []            
            
        return numeric_key,text_key
            
    def _check_cols(self, df, cols):
        """
        Check if required columns are present
        """
        if not bg.tools.isin(a=cols, b=df.columns, how="all", return_element_wise=False):
            raise Exception(f"Some cols are not found in data. \nPresent columns: {df.columns.tolist()}. \nNecessary columns {cols}")


    def _prepare_transition_matrix(self,
                                   df_from,
                                   df_to,
                                   numeric_key,
                                   text_key,
                                   fill_value=0):
        """
        Prepare an empty transition matrix
        """
        # Generate unique
        df_unique_from = df_from[numeric_key+text_key].drop_duplicates()
        df_unique_to = df_to[numeric_key+text_key].drop_duplicates()
        
        # Empty transition matrix
        transition_matrix = pd.DataFrame(data=fill_value,
                                         index=pd.MultiIndex.from_frame(df=df_unique_from),
                                         columns=pd.MultiIndex.from_frame(df=df_unique_to))
        
        return transition_matrix
        
    def _prepare_crosswalk(self,
                           df,
                           numeric_key,
                           text_key):
        """
        Prepare an empty crosswalk dataframe
        """
        # Generate unique
        df_unique = df[numeric_key+text_key].drop_duplicates()
        
        # Empty transition matrix
        crosswalk = pd.DataFrame(index=pd.MultiIndex.from_frame(df=df_unique,names=[f"FROM {c}" for c in df_unique.columns]),
                                 columns=[f"TO {c}" for c in df_unique.columns])

        crosswalk = pd.DataFrame(index=pd.MultiIndex.from_frame(df=df_unique),
                                 columns=df_unique.columns)
        
        return crosswalk
    

    
    
    def _compute_numeric_transition_prob(self,keys,transition_matrix):
        """
        Estimate transition probability matrix from numeric key
        """
        # Copy transition matrix
        P = transition_matrix.copy()
        P_temp = transition_matrix.copy()
        
        if keys:
            # Instantiate 
            numericmatcher = NumericMatcher()
   

            for key in keys:
                
                # Get unique keys to be matched                
                key_from = P.index.get_level_values(level=key).tolist()
                key_to = P.columns.get_level_values(level=key).tolist()
            
                for k in key_from:
                    P_temp[P_temp.index.get_level_values(key) == k] = numericmatcher.compute_similarity(a=k, b=key_to)
                 
                # Add to transition matrix
                P += P_temp
                
            # Normalize total to maximum 1
            P /= len(keys)
        
        return P
        
    def _compute_text_transition_prob(self,keys,transition_matrix):
        """
        Estimate transition probability matrix from text key
        """
        # Copy transition matrix
        P = transition_matrix.copy()
        P_temp = transition_matrix.copy()
        
        if keys:
            # Instantiate 
            textmatcher = TextMatcher()
   
            for key in keys:
                
                # Get unique keys to be matched                
                key_from = P.index.get_level_values(level=key).tolist()
                key_to = P.columns.get_level_values(level=key).tolist()
            
                for k in key_from:
                    P_temp[P_temp.index.get_level_values(key) == k] = textmatcher.compute_similarity(a=k, b=key_to)
                 
                # Add to transition matrix
                P += P_temp
                
            # Normalize total to maximum 1
            P /= len(keys)
        
        return P
                

    def _estimate_transition_matrix(self,
                                    numeric_key,
                                    text_key,
                                    transition_matrix_null):
        
        
        """
        Helper function to estimate transition matrices
        """
        n_transition_matrix = 0
        # Count text_key twice because we use this for context embeddings as well
        for mat in [numeric_key,text_key,text_key]:
            if mat:
                n_transition_matrix += 1
                
        # -------------------------------------------------------------
        # Transition probability matrices
        # -------------------------------------------------------------
        if self.verbose>1:
            bg.tools.print2("Estimating NUMERIC transition matrix")
        
        transition_matrix_numeric = self._compute_numeric_transition_prob(keys=numeric_key,
                                                                                transition_matrix=transition_matrix_null)

        if self.verbose>1:
            bg.tools.print2("Estimating TEXT transition matrix")    
        transition_matrix_text = self._compute_text_transition_prob(keys=text_key,
                                                                          transition_matrix=transition_matrix_null)
 
        if self.verbose>1:
            bg.tools.print2("Estimating CONTEXT transition matrix")
        transition_matrix_context = transition_matrix_null.copy() # Embedding similarity
         
        # Sanity check
        # if transition_matrix_numeric.index != transition_matrix_numeric.index
        
        
        # Combine transition matrices
        transition_matrix_average = transition_matrix_numeric + transition_matrix_text + transition_matrix_context
        
        # Normalize
        transition_matrix_average /= n_transition_matrix
        
        transition_matrix = {
            "numeric" : transition_matrix_numeric,
             "text" : transition_matrix_text,
             "context" : transition_matrix_context,
             "average" : transition_matrix_average
             }
        
        return transition_matrix

    def _find_exact_match(self,transition_matrix):
        
        """
        Find exact matches
        """            
        if not isinstance(transition_matrix,dict):
            raise bg.exceptions.WrongInputTypeException(input_name="transition_matrix",
                                                        provided_input=transition_matrix,
                                                        allowed_inputs=dict)

        # Find exact matches
        mask_exact_numeric = transition_matrix["numeric"] >= 1-self.TOL
        mask_exact_text = transition_matrix["text"] >= 1-self.TOL
        
        mask_exact = mask_exact_numeric | mask_exact_text
        
        # Construct temporary matrix of exact matches (a series of dataframes)
        temp = mask_exact.apply(lambda x: x.index[x == True].to_frame(index=False), axis=1)
        
        # Find empty dataframes (those that do not match exactly)
        mask_empty = pd.Series(data=[temp.loc[idx].empty for idx in temp.index],
                               index=temp.index)
        
        # Subset temporaty
        temp = temp.loc[~mask_empty]
        
        # Turn series of dataframes into dataframe by concatenating
        exact = pd.concat(objs=temp.tolist())

        # Fix index
        exact.index = temp.index

        return exact

    def _find_nbest_match(self,transition_matrix,n=5):
    
        if not isinstance(transition_matrix,pd.DataFrame):
            raise bg.exceptions.WrongInputTypeException(input_name="transition_matrix",
                                                        provided_input=transition_matrix,
                                                        allowed_inputs=pd.DataFrame)
    
    
        # Find n closest numeric matches
        temp = transition_matrix.apply(lambda s, n: s.nlargest(n).index.to_frame(index=False),
                                       axis=1,
                                       n=n)

        nbest = pd.concat(objs=temp.tolist())

        # Fix index by repeating 
        nbest.index = temp.index.repeat(repeats=n)
        
        return nbest
    
    
    def _transpose_transition_matrix(self,transition_matrix):
        
        transition_matrix_transposed = {k:v.T for k,v in transition_matrix.items()}
        
        return transition_matrix_transposed
         
    # -------------------------------------------------------------------------
    # Public functions
    # -------------------------------------------------------------------------
    def estimate_transition_matrix(self,
                                   df_from,
                                   df_to,
                                   use_existing_transition_matrix=True,
                                   numeric_key=None,
                                   numeric_type=None,
                                   text_key=None):
        """
        Estimate transition probability
        
        Note that if 'use_existing_transition_matrix=True', it means that we OVERWRITE existing transition matrices if they do not have ALL unique keys
        """
        
        # ---------------------------------------------------------------------
        # Sanity checks
        # ---------------------------------------------------------------------
        numeric_key,text_key = self._check_keys(numeric_key=numeric_key,
                                                numeric_type=numeric_type,
                                                text_key=text_key)
        
        self._check_cols(df=df_from,cols=numeric_key+text_key)
        self._check_cols(df=df_to,cols=numeric_key+text_key)

        # ---------------------------------------------------------------------
        # Prepare
        # ---------------------------------------------------------------------
        # Prepare a standard transition matrix
        transition_matrix_null = self._prepare_transition_matrix(df_from=df_from,
                                                                 df_to=df_to,
                                                                 numeric_key=numeric_key,
                                                                 text_key=text_key,
                                                                 fill_value=0)
        
        if use_existing_transition_matrix:
            if self.verbose:
                bg.tools.print2("Using existing transition matrix if all keys are present")
                        
            # If we are using an existing transition matrix, we need to make sure that all unique keys are present
            # If not all keys are present, we overwrite exisiting transition matrix
            
            # Get current unique from and to
            unique_from = transition_matrix_null.index
            unique_to = transition_matrix_null.columns

            # Check if they are present in existing
            has_from_keys = all(k in self.unique_from for k in unique_from)
            has_to_keys = all(k in self.unique_to for k in unique_to)                
        
            if (has_from_keys and has_to_keys):
                self.has_all_keys = True
            else:
                self.has_all_keys = False
                

            if not self.has_all_keys:
                
                if self.verbose:
                    bg.exceptions.print_warning("Some keys are missing. Re-estimating transition matrix and overwriting existing")
                
                # Some keys are missing. Hence, we estimate transition matrices again
                self.transition_matrix = self._estimate_transition_matrix(numeric_key=numeric_key,
                                                                          text_key=text_key,
                                                                          transition_matrix_null=transition_matrix_null)

                # Overwrite keys
                self.unique_from = unique_from
                self.unique_to = unique_to

            # Get transition matric                
            transition_matrix = self.transition_matrix
                
        else:
            # Estimate new probability matrix
            transition_matrix = self._estimate_transition_matrix(numeric_key=numeric_key,
                                                                 text_key=text_key,
                                                                 transition_matrix_null=transition_matrix_null)
            


                
        return transition_matrix
                
    
    
    def generate_crosswalk(self,
                           df_from,
                           df_to,
                           numeric_key=None,
                           numeric_type=None,
                           text_key=None):
        """
        Prepare proper crosswalk file
        """
        
        # ---------------------------------------------------------------------
        # Sanity checks
        # ---------------------------------------------------------------------
        numeric_key,text_key = self._check_keys(numeric_key=numeric_key,
                                                numeric_type=numeric_type,
                                                text_key=text_key)
        
        self._check_cols(df=df_from,cols=numeric_key+text_key)
        self._check_cols(df=df_to,cols=numeric_key+text_key)

        # ---------------------------------------------------------------------
        # Estimate transition proabability matrix
        # ---------------------------------------------------------------------
        transition_matrix = self.estimate_transition_matrix(df_from=df_from,
                                                            df_to=df_to,
                                                            use_existing_transition_matrix=True,
                                                            numeric_key=numeric_key,
                                                            numeric_type=numeric_type,
                                                            text_key=text_key)
                
        # ---------------------------------------------------------------------
        # Make crosswalk
        # ---------------------------------------------------------------------
        # Pre-allocate list of crosswalks to be 
        crosswalks = []
        
        crosswalk = self._prepare_crosswalk(df=df_from,
                                            numeric_key=numeric_key,
                                            text_key=text_key)
        
        if self.prioritize_exact_match:
            # Find exact matches
            crosswalk_exact = self._find_exact_match(transition_matrix=transition_matrix)
            
            # Update existing crosswalk
            crosswalk.update(other=crosswalk_exact)
            
            crosswalks.append(crosswalk_exact)
            

        ## Find nbest matches
        # Mask those not found by now
        mask_find = crosswalk.isna().any(axis=1)
        mask_lookup = transition_matrix["average"].index.isin(crosswalk.loc[mask_find].index)

        # Find nbest candidates
        crosswalk_nbest = self._find_nbest_match(transition_matrix=transition_matrix["average"].loc[mask_lookup],
                                                 n=self.n_best_match)

        crosswalks.append(crosswalk_nbest)


        if self.enforce_completeness:
            
            # Transpose all transition matrices
            transition_matrix_to = self._transpose_transition_matrix(transition_matrix=transition_matrix)
            
            # Make sure that all keys in "df_to" are also matched
            crosswalk_to = self._prepare_crosswalk(df=df_to,
                                                   numeric_key=numeric_key,
                                                   text_key=text_key)
                
            # Find exact matches
            crosswalk_exact_to = self._find_exact_match(transition_matrix=transition_matrix_to)
            
            # Update existing crosswalk
            crosswalk_to.update(other=crosswalk_exact_to)
            
            mask_find = crosswalk_to.isna().any(axis=1)
            mask_lookup = transition_matrix_to["average"].index.isin(crosswalk_to.loc[mask_find].index)

            # Find nbest candidates
            crosswalk_nbest_to_temp = self._find_nbest_match(transition_matrix=transition_matrix_to["average"].loc[mask_lookup],
                                                     n=self.n_best_match)

            # Flip index and columns (not transport, but just swap)
            crosswalk_nbest_to = crosswalk_nbest_to_temp.index.to_frame(index=False) 
            crosswalk_nbest_to.index = pd.MultiIndex.from_frame(df=crosswalk_nbest_to_temp)

            crosswalks.append(crosswalk_nbest_to)
            
        # Collect all crosswalks
        df_crosswalk = pd.concat(objs=crosswalks) 

        # Update names
        df_crosswalk.columns = [f"NEW {c}" for c in df_crosswalk.columns]
        
        # Reset index
        df_crosswalk.reset_index(inplace=True)
        
        # Drop duplicates
        df_crosswalk.drop_duplicates(inplace=True)
        
        # Reindex
        df_crosswalk.set_index(keys=numeric_key+text_key, inplace=True)
        

        return df_crosswalk
        
        
    def perform_crosswalk(self,
                          crosswalk,
                          df,
                          values,
                          prefix="NEW ",
                          by=None):
        """
        Perform crosswalk
        """
        # Sanity check
        values = self.__fix_values(v=values)
        self._check_cols(df=df, cols=values)
        
        # Find columns to merge by
        merge_cols = list(crosswalk.index.names)
        merge_cols.sort()
        
        # Fix by
        by = self.__fix_by(by=by,cols=merge_cols)
        
        # Initial merge 
        df_merged = df.merge(right=crosswalk,
                             how='left',
                             on=merge_cols,
                             left_index=False,
                             right_index=False,
                             sort=False,
                             suffixes=('_x', '_y'),
                             copy=True,
                             indicator=False,
                             validate=None)

        # Drop merged cols
        df_merged.drop(columns=merge_cols,
                       inplace=True)

        # Rename
        df_merged.columns = [re.sub(pattern=prefix, repl="", string=c) for c in df_merged.columns]

        # Aggregate      
        df_agg = df_merged.groupby(by=by,as_index=False,sort=True)[values].mean()


        return df_agg
        
        
        
        
        
        

        